/*
*
* SRT2STL
*
*  this prog convert subtitle text generated by subrip (*.srt)
*  for use with spruce DVD maestro (*.stl)
*
* (c) 2000, FredSky
* fredsky@netcourrier.com
*
*/




#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>



int FPS=30;
const maxchar=13;

typedef struct
	{
	int hour;
	int min;
	int sec;
	int msec;
	} MyTime;



/*
* convert time to string in order to display it
* hh:mm:ss:zz
*/
char *timeToString1(MyTime time)
{
	char *res=calloc(maxchar,sizeof(char));
  char *temp=calloc(maxchar,sizeof(char));


	if ((time.hour>=0) && (time.hour<10)) strcat(res,"0");
	itoa(time.hour,temp,10);
  strcat(res,temp);
	strcat(res,":");

	if ((time.min>=0) && (time.min<10)) strcat(res,"0");
  itoa(time.min,temp,10);
  strcat(res,temp);
	strcat(res,":");


	if ((time.sec>=0) && (time.sec<10)) strcat(res,"0");
  itoa(time.sec,temp,10);
  strcat(res,temp);
	strcat(res,":");

	if ((time.msec>=0) && (time.msec<10)) strcat(res,"0");
  itoa(time.msec,temp,10);
  strcat(res,temp);

  free(temp);
	return res;
}


/*
* convert time to string in order to display it
* hh:mm:ss,ddd
*/
char *timeToString2(MyTime time)
{
	char *res=calloc(maxchar,sizeof(char));
  char *temp=calloc(maxchar,sizeof(char));


	if ((time.hour>=0) && (time.hour<10)) strcat(res,"0");
	itoa(time.hour,temp,10);
  strcat(res,temp);
	strcat(res,":");

	if ((time.min>=0) && (time.min<10)) strcat(res,"0");
  itoa(time.min,temp,10);
  strcat(res,temp);
	strcat(res,":");


	if ((time.sec>=0) && (time.sec<10)) strcat(res,"0");
  itoa(time.sec,temp,10);
  strcat(res,temp);
	strcat(res,",");

  itoa(time.msec,temp,10);
  strcat(res,temp);

  free(temp);
	return res;
}



/*
* correct time hh:mm:ss:zz with right values
*/
MyTime correctTime1(MyTime time)
{
	MyTime res;
	div_t t;
	res=time;

	//msec
	if (res.msec>FPS)
	{
		t=div(res.msec,FPS);
		res.msec=t.rem;
		res.sec=res.sec+t.quot;
	}

	//sec
	if (res.sec>60)
	{
		t=div(res.sec,60);
		res.sec=t.rem;
		res.min=res.min+t.quot;
	}

	//min
	if (res.min>60)
	{
		t=div(res.min,60);
		res.min=t.rem;
		res.hour=res.hour+t.quot;
	}

	return res;
}


/*
* correct time hh:mm:ss,ddd with right values
*/
MyTime correctTime2(MyTime time)
{
	MyTime res;
	div_t t;
	res=time;

	//msec
	if (res.msec>1000)
	{
		t=div(res.msec,1000);
		res.msec=t.rem;
		res.sec=res.sec+t.quot;
	}

	//sec
	if (res.sec>60)
	{
		t=div(res.sec,60);
		res.sec=t.rem;
		res.min=res.min+t.quot;
	}

	//min
	if (res.min>60)
	{
		t=div(res.min,60);
		res.min=t.rem;
		res.hour=res.hour+t.quot;
	}

	return res;
}



/*
* encode input values to time
*/
MyTime encodeTime(unsigned int hour, unsigned int min, unsigned int sec, unsigned int msec)
{
	MyTime res;

	res.hour=hour;
	res.min=min;
	res.sec=sec;
	res.msec=msec;
	res=correctTime1(res);

	return res;
}


/*
* add two times (time1+time2)
* hh:mm:ss:zz
*/
MyTime addTime(MyTime time1, MyTime time2)
{
	MyTime res;

	res.msec=time1.msec+time2.msec;
	res.sec=time1.sec+time2.sec;
	res.min=time1.min+time2.min;
	res.hour=time1.hour+time2.hour;
	res=correctTime1(res);

	return res;
}


/*
* subt two times (time1-time2)
* hh:mm:ss:zz
*/
MyTime subTime(MyTime time1, MyTime time2)
{
	MyTime res;

	res.msec=time1.msec-time2.msec;
	res.sec=time1.sec-time2.sec;
	res.min=time1.min-time2.min;
	res.hour=time1.hour-time2.hour;
	res=correctTime1(res);

	if (res.msec<0) {res.msec=res.msec+FPS; res.sec--;}
	if (res.sec<0) {res.sec=res.sec+60; res.min--;}
	if (res.min<0) {res.min=res.min+60; res.hour--;}

	return res;
}


/*
* compare two times
* return  0 if time1=time2
* return  1 if time1>time2
* return -1 if time1<time2
*/
char compTime(MyTime time1, MyTime time2)
{
	MyTime t;

	t=subTime(time1,time2);
	if ( (t.hour==0) && (t.min==0) && (t.sec==0) && (t.msec==0) )
		return 0;
	else
		if ( (t.hour>=0) && (t.min>=0) && (t.sec>=0) && (t.msec>=0) )
			return 1;
		else return -1;

}



/*
* convert hh:mm:ss:zz to decimal
* result in FPS
*/
unsigned long encodeToDecimal1(MyTime time)
{
  unsigned long tot=0;
  char *temp=calloc(10,sizeof(char));

  tot=((time.hour*3600+time.min*60+time.sec)*FPS+time.msec)*33.3335;

  //printf("decimal1 %s\n",ultoa(tot,temp,10));

  free(temp);
	return tot;
}



/*
* convert hh:mm:ss,ddd to decimal
* result in msec
*/
unsigned long encodeToDecimal2(MyTime time)
{
  unsigned int tot2=0;
  unsigned long tot,tot0,tot1;
  char *temp=calloc(10,sizeof(char));
/*
  printf("time.hour %i\n",time.hour);
  printf("time.min %i\n",time.min);
  printf("time.sec %i\n",time.sec);
  printf("time.msec %i\n",time.msec);
*/
  tot0=time.hour*3600000;
  tot1=time.min*60000;
  tot2=time.sec*1000+time.msec;
  tot=tot0+tot1+tot2;

  //printf("decimal2 %s\n",ultoa(tot,temp,10));

  free(temp);
  return tot;
}



/*
* convert decimal value to
* hh:mm:ss,ddd
*/
MyTime convertToTime2(unsigned long value)
{
	MyTime res;
	unsigned long total=value;

	res.msec=total % 1000;
	total=total / 1000;
	res.sec=total % 60;
	res.min=(total / 60) % 60;
	res.hour=total / 3600;

	return res;
}

/*
* convert decimal value to
* hh:mm:ss:zz
*/
MyTime convertToTime1(unsigned long value)
{
	MyTime res;
	unsigned long total=value;

	total=total/33.3335;
	res.msec=total % FPS;
	total=total / FPS;
	res.sec=total % 60;
	res.min=(total / 60) % 60;
	res.hour=total / 3600;

	return res;
}


/*
* count token in string
*/
unsigned int countTokenInString(char *text,char *token)
{
	unsigned int count=0;
  unsigned char len=strlen(text);
	char *test=calloc(len,sizeof(char));
  char *save=calloc(len,sizeof(char));

  strcpy(save,text);

	test=strtok(save,token);
	if (test==NULL)
  {
    free(test);free(save);
    return 0;
  }
	else
	{
		while ( test!=NULL)
		{
			test=strtok(NULL,token);
			count++;
		}
  free(test);free(save);
	return count-1;
	}
}


/*
* get 1 char at start position from input text
*/
char *strget(unsigned char start, char *text)
{
	char *temp=calloc(maxchar,sizeof(char));
	char *dest=calloc(maxchar,sizeof(char));
	char *ptrChar=temp;

	strcpy(temp,text); // copy in temp
	ptrChar=ptrChar+start-1; // avoid the first sartth char from temp
	strncpy(dest,ptrChar,1); //copy 1 char in dest

  free(temp);free(ptrChar);
	return dest;
}


/*
* extract some chars from input text
* star <-> end position
*/
char *strext(unsigned char start, unsigned char end,char *text)
{
	char *temp=calloc(maxchar,sizeof(char));
	char *dest=calloc(maxchar,sizeof(char));
	char *ptrChar=temp;

	strncpy(temp,text,end); // copy the first endth char in temp
	ptrChar=ptrChar+start-1; // avoid the first sartth char from temp
	strcpy(dest,ptrChar); //copy result in dest

  free(temp);free(ptrChar);
	return dest;
}


/*
* convert input string to time
* from hh:mm:ss:zz or hh:mm:ss,zzz
* to hh:mm:ss:zz  (form1)
* or to hh:mm:ss,zzz (form2)
*/
MyTime stringToTime(char *text,unsigned char form)
{
  MyTime res;
  unsigned char len=strlen(text);
	char *t=calloc(len,sizeof(char));
	unsigned int temp;
  unsigned long decimal=0;

	strcpy(t,text);

	//get from hh:mm:ss:zz
	if ( (strcmp(strget(3,t),":")==0) && (strcmp(strget(6,t),":")==0)
		&& (strcmp(strget(9,t),":")==0) && (len==11) )
	{
		temp=atoi(strext(1,2,t));
		res.hour=temp;
    temp=atoi(strext(4,5,t));
		res.min=temp;
		temp=atoi(strext(7,8,t));
    res.sec=temp;
		temp=atoi(strext(10,11,t));
		res.msec=temp;
		res=correctTime1(res);

    //convert from  hh:mm:ss:zz to hh:mm:ss,ddd
    if (form==2)
    {
      decimal=encodeToDecimal1(res);
      res=convertToTime2(decimal);
    }


	}
	else //convert from  hh:mm:ss,ddd to hh:mm:ss:zz
		if ( (strcmp(strget(3,t),":")==0) && (strcmp(strget(6,t),":")==0)
			&& ( (strcmp(strget(9,t),".")==0) || (strcmp(strget(9,t),",")==0) )
      && (len==12) )
		{
      temp=atoi(strext(1,2,t));
      res.hour=temp;
      temp=atoi(strext(4,5,t));
		  res.min=temp;
		  temp=atoi(strext(7,8,t));
      res.sec=temp;
	  	temp=atoi(strext(10,12,t));
		  res.msec=temp;
      res=correctTime2(res);

      //convert from  hh:mm:ss,ddd to hh:mm:ss:zz
      if (form==1)
      {
        decimal=encodeToDecimal2(res);
        res=convertToTime1(decimal);
      }



		}
		else
    {
      printf("\nincorrect input string! '%s'\n",t);
    }


  free(t);
	return res;
}

/*
* search for line where appears
* time of form2 (hh:mm:ss,ddd)
*/
unsigned char getLineTime2(char *text)
{
  unsigned char len=strlen(text);
	char *t=calloc(len,sizeof(char));
  unsigned char res;

  strcpy(t,text);
  if ( (countTokenInString(t,":")==4)
       &&((countTokenInString(t,",")==2)||(countTokenInString(t,".")==2))
       && (len>24) )
       res=1;
  else res=0;

  free(t);
  return res;
}


int main (int argc, char *argv[])
{

	FILE *in, *out, *head; // input,output and headline files
  char *inS=calloc(maxchar,sizeof(char));
  char *outS=calloc(maxchar,sizeof(char)); //string name of input and output
  char *headS=calloc(maxchar,sizeof(char));
  char *test=calloc(maxchar,sizeof(char));

  char *L_in=calloc(300,sizeof(char)); //line read from input file
  char *com=calloc(300,sizeof(char)); //temp comment
  char *comment1=calloc(300,sizeof(char)); //1st comment
  char *comment2=calloc(300,sizeof(char)); //2nd comment
  char *string1=calloc(maxchar,sizeof(char)); //string representing 1st time
  char *string2=calloc(maxchar,sizeof(char)); //string representing 2nd time
  char *string3=calloc(maxchar,sizeof(char)); //string representing 3rd time
  char *string4=calloc(maxchar,sizeof(char)); //string representing 4th time

  char *ptr;

	unsigned char len;
	MyTime t1,t2,t3,t4,t5,t6,t7,t8;
  MyTime time1,time2,time3,time4;



	printf("usage:\n srt2stl input[.srt] output[.stl] headline[.txt] PAL[NSTC]\n");
  printf("\nwhith subtitle input (*.srt),\nand subtitle output (*.stl) \n");
  printf("and headline (*.txt) - default file : headline.txt\n");
  printf("and PAL or NSTC frame rate - default : NSTC at 30fps\n\n");
  printf("Example1 : srt2stl input output\n");
  printf("Example2 : srt2stl input output headline\n");
  printf("Example3 : srt2stl input.srt output.stl headline.txt PAL\n\n\n");

	//test the number of arguments
  if ((argc < 3) || (argc > 5))
	{
    printf("2 args mini and 4 args maxi\n");
		return 1;
	}

	/*
	* test if the input file ends with ".srt"
	* if no, add it
	*/
	strcpy(inS,argv[1]);
	len=strlen(inS);
	test=strext(len-3,len,inS);
	if (stricmp(test,".srt")!=0) strcat(inS,".srt");


  /*
	* test if the output file ends with ".stl"
	* if no, add it
	*/
	strcpy(outS,argv[2]);
	len=strlen(outS);
	test=strext(len-3,len,outS);
	if (stricmp(test,".stl")!=0) strcat(outS,".stl");

  /*
	* test if the headline file ends with ".txt"
	* if no, add it
  * default is "headline.txt"
	*/
  if (argc > 4)
  {
	  strcpy(headS,argv[3]);
	  len=strlen(headS);
	  test=strext(len-3,len,headS);
	  if (stricmp(test,".txt")!=0) strcat(headS,".txt");


    //PAL or NTSC ?
    //default is NTSC 30fps
    if (argc == 5)
    {
      if (stricmp(argv[4],"pal")==0) FPS=25;
      if (stricmp(argv[4],"ntsc")==0) FPS=30;
      printf("FPS %i",FPS);
    }

  }
  else
  //default
    strcpy(headS,"headline.txt");





/*
  //some tests
	printf("\n\n");
	t1=encodeTime(1,1,59,70);
	printf("encodetime  %s",timeToString1(t1));

	printf("\n\n");
	t2=correctTime1(t1);
	printf("correctTime1  %s",timeToString1(t2));

  printf("\n\n");
	printf("comptime  %i",compTime(t2,t2));

  printf("\n\n");
	t3=addTime(t2,t2);
	printf("addtime  %s",timeToString1(t3));

  printf("\n\n");
	t4=subTime(t2,t2);
	printf("subtime  %s",timeToString1(t4));

  printf("\n\n");
	t5=stringToTime("00:11:22:33",1);
	printf("stringtotime11  %s",timeToString1(t5));

  printf("\n\n");
	t6=stringToTime("00:01:47:29",2);
	printf("stringtotime12  %s",timeToString2(t6));

  printf("\n\n");
	t7=stringToTime("00:27:28.830",1);
	printf("stringtotime21  %s",timeToString1(t7));

  printf("\nntoken '00:00:00:01' %i\n",countTokenInString("00:00:00:01",":"));

*/


  //input SRT
  if ((in = fopen(inS, "r"))  == NULL)
   {
      printf("\nCannot open input file '%s'\n",inS);
      return 1;
   }
   //output STL
   if ((out = fopen(outS, "w")) == NULL)
   {
      printf("\nCannot open output file '%s'\n",outS);
      return 1;
   }

   //headline TXT
   if ((head = fopen(headS, "r")) == NULL)
   {
      printf("\nCannot open headline file '%s'\n",headS);
      return 1;
   }



  //copy the headline to the begin of output

  while (!feof(head))
  {
   fputc(fgetc(head),out);
  }


  //main loop

  while (!feof(in))
  {

    //search for the first line with correct timecode format
    do
    {
      fgets(L_in,300,in);
    }
    while ( (getLineTime2(L_in) == 0) && (!feof(in)) );
    //printf("Lin1 '%s'\n",L_in);


    //get the left string time of form hh:mm:ss,ddd
    string1=strext(1,12,L_in);
    //convert it to hh:mm:ss:zz
    time1=stringToTime(string1,1);
    string1=timeToString1(time1);

    //get the right string time of form hh:mm:ss,ddd
    ptr=strrchr(L_in,':');
    len=ptr-L_in;
    string2=strext(len-4,len+7,L_in);
    //convert it to hh:mm:ss:zz
    time2=stringToTime(string2,1);
    string2=timeToString1(time2);

    //get the commentary [next line]
    do
    {
      fgets(com,300,in);
      //no '\n' at the end !
      strncat(comment1,com,strlen(com)-1);
      strcat(comment1,"|");
    }
    while ( (strlen(com) > 2) && (com !=NULL) && (!feof(in)) );

    //delete the two '||' at the end
    comment1[strlen(comment1)-2]='\0';
    printf("%s  %s %s\n",string1,string2,comment1);

    //search for the second line with correct timecode format
    do
    {
      fgets(L_in,300,in);
    }
    while ( (getLineTime2(L_in) == 0) && (L_in !=NULL) && (!feof(in)) );
    //printf("Lin2 '%s'\n",L_in);


    //get the left string time of form hh:mm:ss,ddd
    string3=strext(1,12,L_in);
    //convert it to hh:mm:ss:zz
    time3=stringToTime(string3,1);
    string3=timeToString1(time3);

    //get the right string time of form hh:mm:ss,ddd
    ptr=strrchr(L_in,':');
    len=ptr-L_in;
    string4=strext(len-4,len+7,L_in);
    //convert it to hh:mm:ss:zz
    time4=stringToTime(string4,1);
    string4=timeToString1(time4);

    //get the commentary [next line]
    do
    {
      fgets(com,300,in);
      //no '\n' at the end !
      strncat(comment2,com,strlen(com)-1);
      strcat(comment2,"|");
    }
    while ( (strlen(com) > 2) && (com != NULL) && (!feof(in)) );

    //delete the two '||' at the end
    comment2[strlen(comment2)-2]='\0';
    printf("%s  %s %s\n",string3,string4,comment2);



    /**** compare the differents timecodes ****/


    //if time1<time2 and time3=time1 and time4=time2
    if (  (compTime(time1,time2)==-1) && (compTime(time1,time3)==0)
           && (compTime(time2,time4)==0) )

    { // write only time1 and time2 !

      if (strlen(comment1)>1)
      {
        fputs(string1,out);
        fputs("\t,\t",out);
        fputs(string2,out);
        fputs("\t,\t",out);
        fputs(comment1,out);
        fputs("\n",out);
      }
    }


    else
        //if time1<time2<=time3<time4
        if (  (compTime(time1,time2)==-1) && (compTime(time3,time4)==-1)
           && ((compTime(time2,time3)==-1) || (compTime(time2,time3)==0)))

        { // write all times !

          if (strlen(comment1)>1)
          {
            fputs(string1,out);
            fputs("\t,\t",out);
            fputs(string2,out);
            fputs("\t,\t",out);
            fputs(comment1,out);
            fputs("\n",out);
          }
          if (strlen(comment2)>1)
          {
            fputs(string3,out);
            fputs("\t,\t",out);
            fputs(string4,out);
            fputs("\t,\t",out);
            fputs(comment2,out);
            fputs("\n",out);
          }
        }

    strcpy(string1,"");strcpy(string2,"");
    strcpy(string3,"");strcpy(string4,"");

    strcpy(com,"");
    strcpy(comment1,"");strcpy(comment2,"");


  }



  free(string1);free(string2);
  free(string3);free(string4);

  free(com);
  free(comment1);free(comment2);

  free(L_in);
  free(inS);free(outS);free(test);

	fclose(in);
	fclose(out);
  printf("Conversion DONE.\n");
	return 0;
}